﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.Management;
using Microsoft.Win32.SafeHandles;
using System.Runtime.InteropServices;
using System.IO;


namespace PinscapeConfigTool
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }
        private List<DeviceInfo> devices;

        private static List<DeviceInfo> FindDevices()
        {
            // set up an empty return list
            List<DeviceInfo> devices = new List<DeviceInfo>();

            // get the list of devices matching the HID class GUID
            Guid guid;
            HIDImports.HidD_GetHidGuid(out guid);
            IntPtr hdev = HIDImports.SetupDiGetClassDevs(ref guid, null, IntPtr.Zero, HIDImports.DIGCF_DEVICEINTERFACE);

            // set up the attribute structure buffer
            HIDImports.SP_DEVICE_INTERFACE_DATA diData = new HIDImports.SP_DEVICE_INTERFACE_DATA();
            diData.cbSize = Marshal.SizeOf(diData);

            // read the devices in our list
            for (uint i = 0;
                HIDImports.SetupDiEnumDeviceInterfaces(hdev, IntPtr.Zero, ref guid, i, ref diData);
                ++i)
            {
                // get the size of the detail data structure
                UInt32 size = 0;
                HIDImports.SetupDiGetDeviceInterfaceDetail(hdev, ref diData, IntPtr.Zero, 0, out size, IntPtr.Zero);

                // now actually read the detail data structure
                HIDImports.SP_DEVICE_INTERFACE_DETAIL_DATA diDetail = new HIDImports.SP_DEVICE_INTERFACE_DETAIL_DATA();
                diDetail.cbSize = (IntPtr.Size == 8) ? (uint)8 : (uint)5;
                if (HIDImports.SetupDiGetDeviceInterfaceDetail(hdev, ref diData, ref diDetail, size, out size, IntPtr.Zero))
                {
                    // create a file handle to access the device
                    SafeHandle fp = HIDImports.CreateFile(
                        diDetail.DevicePath, HIDImports.GENERIC_READ_WRITE, HIDImports.SHARE_READ_WRITE, 
                        IntPtr.Zero, FileMode.Open, 0, IntPtr.Zero);

                    // read the attributes
                    HIDImports.HIDD_ATTRIBUTES attrs = new HIDImports.HIDD_ATTRIBUTES();
                    attrs.Size = Marshal.SizeOf(attrs);
                    if (HIDImports.HidD_GetAttributes(fp.DangerousGetHandle(), ref attrs))
                    {
                        // read the product name string
                        String name = "<not available>";
                        byte[] nameBuf = new byte[128];
                        if (HIDImports.HidD_GetProductString(fp.DangerousGetHandle(), nameBuf, 128))
                            name = System.Text.Encoding.Unicode.GetString(nameBuf).TrimEnd('\0');

                        // if the vendor and product ID match an LedWiz, and the
                        // product name contains "pinscape", it's one of ours
                        if ((ushort)attrs.VendorID == 0xFAFA
                            && (attrs.ProductID >= 0x00F0 && attrs.ProductID <= 0x00FF)
                            && Regex.IsMatch(name, @"\b(?i)pinscape\b"))
                        {
                            // add the device to our list
                            devices.Add(new DeviceInfo(
                                fp, diDetail.DevicePath, name, attrs.VendorID,
                                attrs.ProductID, attrs.VersionNumber));

                            // the device list object owns the handle nwo
                            fp = null;
                        }

                        // report the results for debugging
                        Console.Out.WriteLine(
                            "Product " + attrs.ProductID.ToString("X4")
                            + ", vendor " + attrs.VendorID.ToString("X4")
                            + ", version " + attrs.VersionNumber.ToString("X4")
                            + ", name " + name);
                    }
                    
                    // done with the file handle
                    if (fp != null)
                        HIDImports.CloseHandle(fp.DangerousGetHandle());
                }
            }

            // return the device list
            return devices;
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            refreshList();
        }

        private void refreshList()
        {
            // remove any current items in the list
            cbDevices.Items.Clear();
            
            // load the device list
            devices = FindDevices();
            if (devices.Count != 0)
            {
                // populate the popup list
                foreach (DeviceInfo d in devices)
                    cbDevices.Items.Add(d);
            }
            else
            {
                // no devices found
                cbDevices.Items.Add("No controllers found");
            }

            // select the first item
            cbDevices.SelectedItem = cbDevices.Items[0];
        }

        private void btnRefresh_Click(object sender, EventArgs e)
        {
            refreshList();
        }

        private void btnClose_Click(object sender, EventArgs e)
        {
            Form1.ActiveForm.Close();
        }

        private void cbDevices_SelectedIndexChanged(object sender, EventArgs e)
        {
            // get the selected item, and see if it's a device - if not,
            // it's just the "no controllers" string
            var sel = cbDevices.SelectedItem;
            bool ena = false;
            if (sel != null && sel.GetType() == typeof(DeviceInfo))
            {
                // enable the controls
                ena = true;
            }

            // enable/disable buttons depending on whether anything's selected
            btnSettings.Enabled = ena;
            btnPlungerCal.Enabled = ena;
            btnExposure.Enabled = ena;
        }

        private DeviceInfo getSelectedDevice()
        {
            // get the selected item
            var sel = cbDevices.SelectedItem;
            if (sel != null && sel.GetType() == typeof(DeviceInfo))
                return (DeviceInfo)sel;
            else
                return null;
        }
        
        private void btnSettings_Click(object sender, EventArgs e)
        {
            DeviceInfo dev = getSelectedDevice();
            if (dev != null)
            {
                // show the dialog
                Form2 dlg = new Form2(dev);
                dlg.ShowDialog();
            }
        }

        private void btnPlungerCal_Click(object sender, EventArgs e)
        {
            DeviceInfo dev = getSelectedDevice();
            if (dev != null)
            {
                // send the plunger calibration request
                if (dev.SpecialRequest(2))
                {
                    // explain what to do next
                    MessageBox.Show("Calibration request sent - the indicator LED on the device "
                        + "should now show solid blue. To calibrate, pull the plunger all the way "
                        + "back, hold it for a moment, then slowly return to the rest position. "
                        + "Do not let it go or push it forward beyond the normal rest position. "
                        + "Calibration mode automatically ends after 15 seconds.");
                }
            }
        }

        // exposure button - bring up the exposure dialog
        private void btnExposure_Click(object sender, EventArgs e)
        {
            DeviceInfo dev = getSelectedDevice();
            if (dev != null)
            {
                // show the dialog
                Form3 dlg = new Form3(dev);
                dlg.ShowDialog();
            }
        }
    }
}
